# -*- coding: utf-8 -*-
#
# Tiedosto miinaharava.py
#
# Tämä ohjelma on tehty osaksi Aalto-yliopiston ohjelmoinnin oppimateriaaliprojektia.
#
# Tekijät: Teemu Koskinen ja Kimmo Kiiski 2010. 
#
# Jos ä- ja ö-merkit eivät näy oikein ja käytät Eclipseä, vaihda Project-valikon 
# Properties-kohdan kautta käytetty merkistökoodaus. Koodauksen voit vaihtaa kohdasta
# "Text file encoding". Oikea merkistö on UTF-8.
import random

class Miinapeli:
    def __init__(self, leveys, korkeus, miinoja):
        """ Tämä metodi alustaa uuden miinaharava-pelin ja luo annettujen mittojen mukaisen
         pelialueen. """
         
        # Luodaan Pelialue-olio
        self.__pelialue = Pelialue(leveys, korkeus, miinoja)

        # Peli on aluksi käynnissä ja siirtoja on tehty 0.
        self.__peli_kaynnissa = True
        self.__siirtoja = 0
        
    
    def kerro_pelialueen_mitat(self):
        """ Hakee pelialueelta sen mitat. Arvo palautetaan monikkona, jonka 
         indeksissä 0 on leveys ja indeksissä 1 korkeus. """
        return self.__pelialue.kerro_mitat()


    def onko_kaynnissa(self):
        """ Palauttaa tiedon siitä, onko peli käynnissä. """
        return self.__peli_kaynnissa        

       
    def anna_ruutu(self, x, y):
        """ Palauttaa annetuissa koordinaateissa sijaitsevan Ruutu-olion pelialueelta. """
        return self.__pelialue.hae_ruutu(x, y)


    def kerro_siirrot(self):
        """ Kertoo pelaajan tekemien siirtojen määrän. """
        return self.__siirtoja
    

    def miinoja_ruudun_vieressa(self, ruutu):
        """ Palauttaa tiedon siitä, kuinka monta miinaa annettujen 
         koordinaattien vieressä on. """
        return self.__pelialue.laske_viereiset_miinat(ruutu)


    def onko_voittanut(self):
        """ Tämä metodi palauttaa tiedon siitä, onko pelaaja voittanut, eli onko 
         hän avannut kaikki miinattomat ruudut. 
         
         Metodi etsii rivi kerrallaan ruutua, jossa ei ole miinaa ja joka
         ei ole auki. Jos avaamaton miinaton ruutu tulee vastaan, 
         palautetaan False. Jos pelialueen kaikki ruudut käydään läpi
         löytämättä yhtään avaamatonta ja miinatonta ruutua, peli on ratkaistu 
         ja metodi palauttaa arvon True. """
        
        leveys, korkeus = self.kerro_pelialueen_mitat()
        for y in range(korkeus):
            for x in range(leveys):
                ruutu = self.anna_ruutu(x, y)
                if not ruutu.onko_auki() and not ruutu.onko_miina():
                    # Peli on vielä kesken joten palautetaan False.
                    return False
                
        # Kaikki ruudut käytiin läpi löytämättä yhtään avaamatonta miinatonta ruutua.
        # Peli on siis ratkaistu.
        return True
        
        
    def tee_siirto(self, ruutu):
        """ Tämä metodi avaa parametriarvona annetun ruudun ja kasvattaa 
         siirtolaskurin arvoa yhdellä. """
              
        # Kutsutaan ruudun avaavaa metodia:
        self.__avaa_ruutu(ruutu)

        # Kasvatetaan siirtolaskurin arvoa yhdellä:
        self.__siirtoja += 1
                        
        
    def __avaa_ruutu(self, ruutu):
        """ Avaa parametriarvona annetun ruudun. Jos ruudussa on 
         miina, peli päättyy. 
         
         Jos ruudun vieressä ei ole lainkaan miinoja, tämä metodi huolehtii 
         siitä, että ruudun naapuriruudut avataan automaattisesti. """
        
        # Jos osuttiin miinaan heti ensimmäisellä siirrolla, siirretään 
        # miina toiseen (satunnaisesti valittavaan) ruutuun.
        if ruutu.onko_miina() and self.kerro_siirrot() == 0:
            self.__pelialue.siirra_miina(ruutu)
        
        # Avataan ruutu
        ruutu.avaa()

        # Tarkastetaan, oliko ruutu miina
        if not ruutu.onko_miina():
            # Jos ruudun vieressä ei ole lainkaan miinoja, avataan myös sen naapurit:
            if self.miinoja_ruudun_vieressa(ruutu) == 0:
                self.__avaa_naapurit(ruutu)
            
            # Tarkistaa onko pelaaja voittanut. Jos pelaaja voitti, päätetään peli.
            if self.onko_voittanut():
                self.__lopeta_peli(True)

        else:
            # Ruutu oli miina, joten peli päättyy
            self.__lopeta_peli(False)
    

    def __avaa_naapurit(self, ruutu):
        """ Tämä metodi avaa annetun koordinaatin viereiset ruudut, jotka eivät ole 
         vielä auki, kutsumalla avaa_ruutu-metodia. """
         
        # Käydään läpi naapuri kerrallaan
        for naapuri in self.__pelialue.hae_naapurit(ruutu):
            # Jos naapuri ei ole auki eikä lippu, avataan se
            if not naapuri.onko_auki() and not naapuri.onko_lippu():
                self.__avaa_ruutu(naapuri)
                    
                
    def __lopeta_peli(self, voittanut):
        """ Lopettaa pelin. Metodi ottaa parametriarvonaan totuusarvon. Jos arvo 
         on True, pelaaja merkitään voittaneeksi pelin. Jos arvo on False, 
         pelaaja hävisi ja pelialueen kaikki miinoja sisältävät ruudut muutetaan
         näkyviksi. """
         
        # Peli on päättynyt
        self.__peli_kaynnissa = False
        
        if not voittanut:
            # Pelaaja hävisi, joten näytetään miinat
            self.__avaa_miinat()
        
    
    def __avaa_miinat(self):
        """ Muuttaa kaikki miinoja sisältävät ruudut avatuiksi. Tätä metodia
         käytetään pelaajan hävitessä pelin, jolloin kaikki miinat tulevat 
         näkyviin. """
        
        # Käydään pelialueen kaikki rivit ja sarakkeet läpi
        leveys, korkeus = self.kerro_pelialueen_mitat()
        for x in range(leveys):
            for y in range(korkeus):
                # Haetaan ruutu, ja jos siinä on miina, avataan se
                ruutu = self.anna_ruutu(x, y)
                if ruutu.onko_miina():
                    ruutu.avaa()

        
    def avaa_viereiset_liputtomat(self, ruutu):
        """ Tällä metodilla voidaan avata kaikki annetun ruudun vieressä sijaitsevat
         liputtamattomat ruudut, jos ruudun vieressä on yhtä monta lippua kuin miinoja. """
        
        
        naapureita_liputettu = self.__pelialue.laske_viereiset_liput(ruutu)

        # Ruudun tulee olla auki ja sen vieressä tulee olla yhtä monta
        # liputettua ruutua, kuin miinoja
        if ruutu.onko_auki() and naapureita_liputettu == self.miinoja_ruudun_vieressa(ruutu):
            
            # Kasvatetaan siirtolaskurin arvoa
            self.__siirtoja += 1
        
            # Käydään kaikki naapurit läpi
            for naapuri in self.__pelialue.hae_naapurit(ruutu):
                # Avataan kaikki naapurit, joita ei ole liputettu ja jotka eivät ole auki
                if not naapuri.onko_lippu() and not naapuri.onko_auki():

                    # Avataan jokainen ruutu vuorollaan __avaa_ruutu-metodilla.
                    self.__avaa_ruutu(naapuri)
    
    
class Pelialue:
    def __init__(self, leveys, korkeus, miinoja):
        """ Alustaa uuden pelialueen annetuilla mitoilla ja miinojen määrällä. """
        self.__leveys = leveys
        self.__korkeus = korkeus
        
        self.__luo_ruudukko() 
        
        self.__aseta_miinat(miinoja)
        
        
    def __luo_ruudukko(self):
        """ Luo ruudukon ja täyttää sen Ruutu-olioilla. 
         Tässä kaksiulotteista ruudukkoa simuloidaan käyttämällä yksiulotteista listaa,
         jossa koordinaateissa x, y sijaitseva Ruutu sijaitsee indeksissä x + y*leveys. """

        # Luodaan lista, jota käytetään ruudukkona
        self.__ruudukko = [] 

        # Luodaan ruudukko-listaan pelialueen koon verran uusia Ruutu-olioita.
        for y in range(self.__korkeus):
            for x in range(self.__leveys):
                # Jokaista x- ja y-koordinaattiparia vasten luodaan uusi Ruutu-olio,
                # joka lisätään ruudukko-listaan. 
                uusi_ruutu = Ruutu(x, y)
                self.__ruudukko.append(uusi_ruutu)
    
    
#    Toinen versio, joka näytetään miinaharava-esitysten osassa 3:
#
#    def __luo_ruudukko(self):
#        """ Luo uuden kaksiulotteisen ruudukkolistan ja täyttää sen
#         Ruutu-olioilla. Kaksiulotteinen lista toteutetaan sisäkkäisillä
#         listoilla, joissa ulommainen lista sisältää jokaisessa
#         indeksissään toisen listan. Sisemmät listat sisältävät varsinaisia
#         Ruutu-olioita. """
#
#        # Luodaan ulommainen lista, joka kuvaa x-akselia.
#        self.__ruudukko = [] 
#
#        # Luodaan ulompaan listaan pelialueen leveyden verran uusia listoja,
#        # joista kuhunkin sijoitetaan pelialueen korkeuden verran Ruutu-olioita. 
#        for x in range(self.__leveys):
#            
#            # Luo uusi sarake
#            uusi_sarake = []
#            
#            # Luodaan sisempään listaan pelialueen korkeuden verran uusia ruutuja
#            for y in range(self.__korkeus):
#                uusi_ruutu = Ruutu(x, y)
#                uusi_sarake.append(uusi_ruutu)
#            
#            # Lisää sarake ruudukkoon
#            self.__ruudukko.append(uusi_sarake)
    
    
    def kerro_mitat(self):
        """ Palauttaa ruudukon mitat eli leveyden ja korkeuden. """
        return self.__leveys, self.__korkeus
    
    
    def __aseta_miinat(self, lukumaara):
        """ Sijoittaa parametriarvona annetun määrän miinoja pelialueelle satunnaisiin 
         ruutuihin. """
         
        """ Huomautus: Koska tämä metodi etsii aina satunnaisia ruutuja miinoitettaviksi, yritetään 
         mahdollisesti samaa ruutua miinoittaa moneen kertaan. Jos miinoja asetettaisiin pelialueen 
         kokoon verrattuna paljon, kestäisi miinojen asettaminen mahdollisesti suhteettoman 
         kauan (pahimmassa tapauksessa tämän metodin suoritus ei koskaan päättyisi). 
         
         Tämän metodin tehokkuutta voitaisiin parantaa esimerkiksi käyttämällä apuna listaa, jossa 
         pidettäisiin kaikkia miinoittamattomia ruutuja. Satunnainen ruutu voitaisiin valita ja poistaa
         tältä apulistalta, jolloin ei olisi pelkoa siitä, että sama ruutu valittaisiin moneen 
         kertaan. """
         
        # Pidetään kirjaa asetettujen ruutujen määrästä
        asetettu = 0
        
        # Toistetaan kunnes haluttu määrä miinoja on asetettu
        while asetettu < lukumaara:
            
            # Otetaan satunnainen ruutu
            ruutu = self.__hae_satunnainen_ruutu()
            
            # Jos ruutu ei ole miina, asetetaan se miinaksi ja päivitetään
            # laskuri.
            if not ruutu.onko_miina():
                ruutu.aseta_miina()
                asetettu += 1
                       

    def hae_ruutu(self, x, y):
        """ Palauttaa annetuissa koordinaateissa sijaitsevan Ruutu-olion. """
        
        # Koordinaatteja x ja y vastaava ruutu sijaitsee indeksissä y*leveys + x.
        return self.__ruudukko[y*self.__leveys + x]
 
    
#    Toinen versio, joka näytetään miinaharava-esitysten osassa 3:
#
#    def hae_ruutu(self, x, y):
#        """ Palauttaa annetuissa koordinaateissa sijaitsevan Ruutu-olion. """
#        
#        # Koordinaatteja x ja y vastaava ruutu sijaitsee indekseissä x ja y.
#        return self.__ruudukko[x][y]

   
    def __hae_satunnainen_ruutu(self):
        """ Palauttaa satunnaisen Ruutu-olion. """
               
        # Arvotaan koordinaatit sopivilta väleiltä
        x = random.randint(0, self.__leveys - 1)
        y = random.randint(0, self.__korkeus - 1)
        
        # Haetaan ja palautetaan koordinaateissa sijaitseva ruutu
        return self.hae_ruutu(x, y)
    
    
    def siirra_miina(self, ruutu):
        """ Poistaa miinan parametrina annetusta ruudusta, ja lisää miinan 
         satunnaiseen ruutuun, joka ei voi olla sama kuin parametrina annettu, eikä
         sisältää valmiiksi miinaa. """
         
        # Asetetaan ensin annettu ruutu uudeksi ruuduksi. 
        # Tämä ruutu ei ole kelvollinen, jos siinä on jo miina, joten
        # suoritus etenee aluksi while-toistokäskyyn.
        uusi_ruutu = ruutu
                
        # Arvotaan uusi ruutu, kunnes löydetään ruutu, joka 
        # ei ole valmiiksi miinoitettu.
        while uusi_ruutu.onko_miina():
            uusi_ruutu = self.__hae_satunnainen_ruutu()

        # Poistetaan miina parametrina annetusta ruudusta 
        ruutu.poista_miina()   
             
        # Asetetaan miina uuteen, arvottuun ruutuun.
        uusi_ruutu.aseta_miina()
    
    
    
    def hae_ruutu_kaksiulotteiselta_listalta(self, x, y):
        """ Tämä metodi on vaihtoehtoinen toteutus, jota voidaan 
         käyttää kaksiulotteisten listojen kanssa. """
         
        """ Palauttaa annetuissa koordinaateissa sijaitsevan Ruutu-olion. """
        return self.__ruudukko[x][y]
        
    
    def hae_naapurit(self, ruutu):
        """ Palauttaa listan niistä ruuduista, jotka sijaitsevat annetuissa
         koordinaateissa olevan ruudun vieressä. Myös kulmistaan vierekkäin olevat
         naapurit lisätään listaan. """

        x, y = ruutu.kerro_sijainti()
        naapurit = []
                
        # Käydään läpi vaaka- ja pystysuunnassa olevat viereiset koordinaatit.
        for x_koord in (x-1, x, x+1):
            for y_koord in (y-1, y, y+1):
                # Tarkastetaan että ruutu ei ole sama, jonka naapureita etsitään,
                # sekä se, että se ei sijaitse pelialueen ulkopuolella.
                if ( self.onko_ruutu_kelvollinen(x_koord, y_koord) and
                        not (x_koord == x and y_koord == y)):
                    # Haetaan viereisessä ruudussa oleva Ruutu-olio ja
                    # lisätään lisätään se listaan.
                    naapuri = self.hae_ruutu(x_koord, y_koord)
                    naapurit.append(naapuri)
        
        return naapurit


    def laske_viereiset_liput(self, ruutu):
        """ Palauttaa tiedon siitä, kuinka monta annetun ruudun naapuria on
         merkitty lipulla. """
        lippuja = 0
        
        # Käydään läpi ruudun naapurit, ja tarkastetaan ovatko ne liputettuja
        naapurit = self.hae_naapurit(ruutu)
        for naapuri in naapurit:
            if naapuri.onko_lippu():
                lippuja += 1
        
        # Palautetaan löydettyjen lippujen summa
        return lippuja
    
    
    def laske_viereiset_miinat(self, ruutu):
        """ Palauttaa tiedon siitä, kuinka monta ruudun naapuria on miinoitettu. """
        miinoja = 0
        
        # Käydään läpi ruudun naapurit, ja tarkastetaan ovatko ne miinoja
        naapurit = self.hae_naapurit(ruutu)
        for naapuri in naapurit:
            if naapuri.onko_miina():
                miinoja += 1
        
        # Palautetaan löydettyjen miinojen summa        
        return miinoja
    
    
    def onko_ruutu_kelvollinen(self, x, y):
        """ Palauttaa tiedon siitä, sijaitsevatko annetut
         koordinaatit pelialueella. """
        return (0 <= x < self.__leveys and 0 <= y < self.__korkeus)

    
    def __str__(self):
        """ Muodostaa ja palauttaa tekstimuotoisen kuvauksen ruudukon 
         sisällöstä. ___str___ metodilla voidaan yleisesti muodostaa olioista
         kuvauksia, jotka esimerkiksi helpottavat virheiden etsintää. """
        
        kuvaus = ""
        # Käydään ruudukko läpi rivi kerrallaan
        for y in range(self.__korkeus):
            # Jokaisesta sarakkeesta lisätään kuvaukseen merkki kuvaamaan
            # siinä sijaitsevaa ruutua. 
            for x in range(self.__leveys):
                ruutu = self.hae_ruutu(x, y)
                if ruutu.onko_miina():
                    kuvaus += "¤"
                else:
                    miinoja_vieressa = self.laske_viereiset_miinat(ruutu) 
                    if miinoja_vieressa == 0:
                        kuvaus += "-"
                    else: 
                        kuvaus += str(miinoja_vieressa)
                
                # Saman rivin Ruudut erotellaan toisistaan välilyönneillä.
                kuvaus += " "   
            
            # Lisätään rivinvaihto jokaisen rivin jälkeen.
            kuvaus += "\n"
            
        # Palautetaan muodostettu kuvaus
        return kuvaus


class Ruutu:
    def __init__(self, x, y):
        """ Alustaa uuden Ruutu-olion. Jokainen ruutu on aluksi tyhjä ja suljettu
         ja sen vieressä ei ole miinoitettuja ruutuja. """
        self.__on_miina = False
        self.__on_auki = False
        self.__on_lippu = False
        
        # Tallennetaan koordinaatit
        self.__x = x
        self.__y = y
    
    
    def onko_lippu(self):
        """ Palauttaa tiedon, onko ruutu merkitty lipuksi. """
        return self.__on_lippu
    
    
    def aseta_lippu(self):
        """ Asettaa ruudun liputetuksi. """
        self.__on_lippu = True


    def poista_lippu(self):
        """ Poistaa lipun ruudusta. """
        self.__on_lippu = False
                
    
    def onko_miina(self):
        """ Palauttaa tiedon siitä, onko ruutu miinoitettu. """
        return self.__on_miina
    
            
    def aseta_miina(self):
        """ Asettaa ruutuun miinan. """
        self.__on_miina = True

        
    def poista_miina(self):
        """ Poistaa ruudusta miinan. """
        self.__on_miina = False
            
    
    def onko_auki(self):
        """ Palauttaa tiedon siitä, onko ruutu avattu. """
        return self.__on_auki
    
            
    def avaa(self):
        """ Avaa ruudun. """
        self.__on_auki = True
    
        
    def kerro_sijainti(self):
        """ Palauttaa Ruutu-olion koordinaatit pelialueella monikkona
         muodossa (x-koordinaatti, y-koordinaatti). """
        return (self.__x, self.__y)


    def __str__(self):
        """ Palauttaa merkkijonomuotoisen kuvauksen oliosta. 
         Jos Ruutu-olio on merkitty miinaksi, kuvauksena käytetään ¤-merkkiä. 
         Muussa tapauksessa palautetaan yhdysviiva-merkki. """
        if self.onko_miina():
            return "¤"
        else: 
            return "-"

def testi():
    """ Luodaan uusi pelialue, jonka koko on 10*5 ruutua ja siinä on 3 miinaa. """
    alue1 = Pelialue(10, 5, 3)

    """ Tulostetaan pelialueen kuvaus hyödyntäen Pelialue-luokan __str__-metodia. """
    print(alue1)

    alue2 = Pelialue(8, 8, 5)
    print(alue2)
    
    alue3 = Pelialue(5, 7, 4)
    print(alue3)

    
    
""" Tarkastetaan, suoritettiinko ohjelma alkaen tästä moduulista vai
 ollaanko tätä moduulia tuomassa import-käskyllä toiseen moduuliin.

 Se moduuli, josta ohjelma käynnistetään saa nimekseen aina
 automaattisesti __main__. Suoritettavan moduulin nimen saa puolestaan
 selville muuttujan __name__ avulla. """
if __name__ == "__main__":
    # Suoritus alkoi tästä moduulista, joten kutsutaan testi-funktiota.
    testi()
